# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

"""Helper functions for the Bayesian bootstrap."""

from __future__ import annotations

from collections.abc import Callable

import numpy as np
from numpy.typing import NDArray

from samesame._parallel import Resampler, _simulate_in_parallel


def _rudirichlet(
    n_rows: int,
    n_resamples: int = 9999,
    rng: np.random.Generator = np.random.default_rng(),
) -> NDArray[np.float64]:
    """from R function:
    https://github.com/vathymut/dsos/blob/53c8e7bb5e1aef9bd093f9f7b6bf2fbb36494de5/R/bayes-factor.R#L3
    Also see function in bayesboot package
    """
    weights = rng.standard_exponential(size=(n_rows, n_resamples))
    weights /= weights.sum(axis=0)
    return weights * n_rows


def _bayesian_bootstrap(
    fn: Callable[[NDArray], float],
    n_rows: int,
    n_resamples: int = 9999,
    rng: np.random.Generator = np.random.default_rng(),
) -> NDArray[np.float64]:
    weights = _rudirichlet(n_rows=n_rows, n_resamples=n_resamples, rng=rng)
    return np.apply_along_axis(fn, 0, weights)


def _bayesian_posterior(
    actual: NDArray[np.int_],
    predicted: NDArray,
    metric: Callable,
    n_resamples: int = 9999,
    rng: np.random.Generator = np.random.default_rng(),
):
    def fn(sample_weight):
        return metric(actual, predicted, sample_weight=sample_weight)

    n_rows = len(actual)
    return _bayesian_bootstrap(
        fn=fn,
        n_rows=n_rows,
        n_resamples=n_resamples,
        rng=rng,
    )


def bayesian_posterior(
    actual: NDArray,
    predicted: NDArray,
    metric: Callable,
    n_resamples: int = 9999,
    rng: np.random.Generator = np.random.default_rng(),
    n_jobs: int = 1,
    batch: int | None = None,
):
    resampler = Resampler(
        n_resamples=n_resamples,
        rng=rng,
        n_jobs=n_jobs,
        batch=batch,
    )

    def _simulate_fn(rng=rng, n_resamples=n_resamples):
        return _bayesian_posterior(
            actual=actual,
            predicted=predicted,
            metric=metric,
            n_resamples=n_resamples,
            rng=rng,
        )

    return _simulate_in_parallel(
        resampler=resampler,
        simulate_fn=_simulate_fn,
    )



# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
from dataclasses import dataclass
from functools import cached_property

import numpy as np
from numpy.typing import NDArray
from sklearn.utils.multiclass import type_of_target

from samesame._bayesboot import bayesian_posterior
from samesame._data import assign_labels, concat_samples
from samesame._stats import _bayes_factor
from samesame.ctst import CTST
from samesame.metrics import wauc


@dataclass
class WeightedAUC(CTST):
    """
    Two-sample test for no adverse shift using the weighted AUC (WAUC).

    This test compares scores from two independent samples. We reject the
    null hypothesis of no adverse shift for unusually high values of the WAUC
    i.e. when the second sample is relatively worse than the first one. This
    is a robust nonparametric noninferiority test (NIT) with no pre-specified
    margin. It can be used, amongst other things, to detect dataset shift with
    outlier scores, hence the DSOS acronym.

    Attributes
    ----------
    actual : NDArray
        Binary indicator for sample membership.
    predicted : NDArray
        Estimated (predicted) scores for corresponding samples in `actual`.
    n_resamples : int, optional
        Number of resampling iterations, by default 9999.
    rng : np.random.Generator, optional
        Random number generator, by default np.random.default_rng().
    n_jobs : int, optional
        Number of parallel jobs, by default 1.
    batch : int or None, optional
        Batch size for parallel processing, by default None.

    See Also
    --------
    bayes.as_bf : Convert a one-sided p-value to a Bayes factor.

    bayes.as_pvalue : Convert a Bayes factor to a one-sided p-value.

    Notes
    -----
    The frequentist null distribution of the WAUC is based on permutations
    [1]. The Bayesian posterior distribution of the WAUC is based on the
    Bayesian bootstrap [2]. Because this is a one-tailed test of direction
    (it asks the question, 'are we worse off?'), we can convert a one-sided
    p-value into a Bayes factor and vice versa. We can also use these p-values
    for sequential testing [3].

    The test assumes that `predicted` are outlier scores and/or encode some
    notions of outlyingness; higher value of `predicted` indicates worse
    outcomes.

    References
    ----------
    .. [1] Kamulete, Vathy M. "Test for non-negligible adverse shifts."
       Uncertainty in Artificial Intelligence. PMLR, 2022.

    .. [2] Gu, Jiezhun, Subhashis Ghosal, and Anindya Roy. "Bayesian bootstrap
       estimation of ROC curve." Statistics in medicine 27.26 (2008): 5407-5420.

    .. [3] Kamulete, Vathy M. "Are you OK? A Bayesian Sequential Test for
       Adverse Shift." 2025.

    Examples
    --------
    >>> import numpy as np
    >>> from samesame.nit import WeightedAUC
    >>> # alternatively: from samesame.nit import DSOS
    >>> actual = np.array([0, 1, 1, 0])
    >>> scores = np.array([0.2, 0.8, 0.6, 0.4])
    >>> wauc = WeightedAUC(actual, scores)
    >>> print(wauc.pvalue) # doctest: +SKIP
    >>> print(wauc.bayes_factor) # doctest: +SKIP
    >>> wauc_ = WeightedAUC.from_samples(scores, scores)
    >>> isinstance(wauc_, WeightedAUC)
    True
    """

    def __init__(
        self,
        actual: NDArray,
        predicted: NDArray,
        n_resamples: int = 9999,
        rng: np.random.Generator = np.random.default_rng(),
        n_jobs: int = 1,
        batch: int | None = None,
    ):
        """Initialize WeightedAUC with fixed metric and alternative."""
        super().__init__(
            actual=actual,
            predicted=predicted,
            metric=wauc,
            n_resamples=n_resamples,
            rng=rng,
            n_jobs=n_jobs,
            batch=batch,
            alternative="greater",
        )

    @cached_property
    def posterior(self) -> NDArray:
        """
        Compute the posterior distribution of the WAUC.

        Returns
        -------
        NDArray
            The posterior distribution of the WAUC.

        Notes
        -----
        The result is cached to avoid (expensive) recomputation since the
        posterior distribution uses the Bayesian bootstrap.
        """
        return bayesian_posterior(
            self.actual,
            self.predicted,
            self.metric,
            self.n_resamples,
            self.rng,
            self.n_jobs,
            self.batch,
        )

    @cached_property
    def bayes_factor(self):
        """
        Compute the Bayes factor using the Bayesian bootstrap.

        Notes
        -----
        The result is cached to avoid (expensive) recomputation.
        """
        bayes_threshold = float(np.mean(self.null))
        bf_ = _bayes_factor(self.posterior, bayes_threshold)
        return bf_

    @classmethod
    def from_samples(
        cls,
        first_sample: NDArray,
        second_sample: NDArray,
        n_resamples: int = 9999,
        rng: np.random.Generator = np.random.default_rng(),
        n_jobs: int = 1,
        batch: int | None = None,
    ):
        """
        Create a WeightedAUC instance from two samples.

        Parameters
        ----------
        first_sample : NDArray
            First sample of scores. These can be binary or continuous.
        second_sample : NDArray
            Second sample of scores. These can be binary or continuous.

        Returns
        -------
        WeightedAUC
            An instance of the WeightedAUC class.
        """
        assert type_of_target(first_sample) == type_of_target(second_sample)
        samples = (first_sample, second_sample)
        actual = assign_labels(samples)
        predicted = concat_samples(samples)
        return cls(
            actual,
            predicted,
            n_resamples,
            rng,
            n_jobs,
            batch,
        )


DSOS = WeightedAUC

__all__ = [
    "WeightedAUC",
    "DSOS",
]



# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
"""Compute the weighted area under the ROC."""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from scipy.integrate import trapezoid
from sklearn.metrics import roc_curve

from samesame._data import group_by
from samesame._wecdf import ECDFDiscrete


def wauc(
    actual: NDArray[np.int_],
    predicted: NDArray,
    *,
    sample_weight: NDArray | None = None,
) -> float:
    """
    Compute the weighted area under the ROC curve (WAUC).

    Calculates the WAUC by weighting the true positive rate (TPR) at each
    false positive rate (FPR) threshold, optionally using sample weights.
    The weights are computed as the squared empirical weighted cumulative
    distribution function (EW-CDF) of the predicted scores for the negative
    class.

    Parameters
    ----------
    actual : NDArray[np.int_]
        Ground truth binary labels (0 or 1).
    predicted : NDArray
        Predicted scores or probabilities.
    sample_weight : NDArray or None, optional
        Sample weights. If None, all samples are given equal weight.

    Returns
    -------
    float
        The weighted area under the ROC curve.

    Notes
    -----
    The function uses the `roc_curve` from scikit-learn to compute FPR, TPR,
    and thresholds. The empirical weighted CDF is computed for the negative
    class predictions using `ECDFDiscrete`. The WAUC is calculated using the
    trapezoidal rule, weighting the TPR by the squared EW-CDF at each
    threshold [1].

    References
    ----------
    .. [1] Li, Jialiang, and Jason P. Fine. "Weighted Area under the Receiver
       Operating Characteristic Curve and Its Application to Gene Selection."
       Journal of the Royal Statistical Society: Series C (Applied Statistics),
       vol. 59, no. 4, 2010, pp. 673-692.

    Examples
    --------
    >>> import numpy as np
    >>> from samesame.metrics import wauc
    >>> actual = np.array([0, 1, 0, 1])
    >>> predicted = np.array([0.1, 0.4, 0.35, 0.8])
    >>> wauc(actual, predicted)
    np.float64(0.625)
    """
    fpr, tpr, thresholds = roc_curve(
        actual,
        predicted,
        pos_label=None,
        sample_weight=sample_weight,
    )
    if sample_weight is not None:
        stacked = np.column_stack((predicted, sample_weight))
        grp_0 = group_by(data=stacked, groups=actual)[0]
        ewcdf = ECDFDiscrete(grp_0[:, 0], freq_weights=grp_0[:, 1])
    else:
        grp_0 = group_by(data=predicted, groups=actual)[0]
        ewcdf = ECDFDiscrete(grp_0)
    weights = np.power(ewcdf(thresholds), 2)
    return trapezoid(
        y=tpr * weights,
        x=fpr,
    )



# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
"""
Classifier two-sample tests (CTST) from binary classification metrics.

The classifier two-sample test broadly consists of three steps: (1)
training a classifier, (2) scoring the two samples and (3) turning a test
statistic into a p-value from these scores. This test statistic can be the
performance metric of a binary classifier such as the (weighted) area under
the receiver operating characteristic curve, the Matthews correlation
coefficient, and the (balanced) accuracy. This module tackles step (3).

References
----------
.. [1] Lopez-Paz, David, and Maxime Oquab. "Revisiting Classifier Two-Sample
   Tests." International Conference on Learning Representations. 2017.

.. [2] Friedman, Jerome. "On multivariate goodness-of-fit and two-sample
   testing." No. SLAC-PUB-10325. SLAC National Accelerator Laboratory (SLAC),
   Menlo Park, CA (United States), 2004.

.. [3] Kübler, Jonas M., et al. "Automl two-sample test." Advances in Neural
   Information Processing Systems 35 (2022): 15929-15941.

.. [4] Ciémençon, Stéphan, Marine Depecker, and Nicolas Vayatis. "AUC
   optimization and the two-sample problem." Proceedings of the 23rd
   International Conference on Neural Information Processing Systems. 2009.

.. [5] Hediger, Simon, Loris Michel, and Jeffrey Näf. "On the use of random
   forest for two-sample testing." Computational Statistics & Data Analysis
   170 (2022): 107435.

.. [6] Kim, Ilmun, et al. "Classification accuracy as a proxy for two-sample
   testing." Annals of Statistics 49.1 (2021): 411-434.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from functools import cached_property
from inspect import signature
from operator import methodcaller
from typing import Literal

import numpy as np
from numpy.typing import NDArray
from sklearn.utils import (
    check_consistent_length,
    column_or_1d,
)
from sklearn.utils.multiclass import type_of_target

from samesame._data import assign_labels, concat_samples
from samesame._permute import permutation_null
from samesame._stats import EmpiricalPvalue
from samesame._utils import check_metric_function


@dataclass
class CTST:
    """
    Classifier two-sample test (CTST) using a binary classification metric.

    This test compares scores (predictions) from two independent samples.
    Rejecting the null implies that scoring is not random and that the
    classifier is able to distinguish between the two samples.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.metrics import balanced_accuracy_score, matthews_corrcoef, roc_auc_score
    >>> from samesame.ctst import CTST
    >>> actual = np.array([0, 1, 1, 0])
    >>> scores = np.array([0.2, 0.8, 0.6, 0.4])
    >>> ctst_acc = CTST(actual, scores, metric=balanced_accuracy_score)
    >>> ctst_mcc = CTST(actual, scores, metric=matthews_corrcoef)
    >>> ctst_auc = CTST(actual, scores, metric=roc_auc_score)
    >>> print(ctst_acc.pvalue) # doctest: +SKIP
    >>> print(ctst_mcc.pvalue) # doctest: +SKIP
    >>> print(ctst_auc.pvalue) # doctest: +SKIP
    >>> ctst_ = CTST.from_samples(scores, scores, metric=roc_auc_score)
    >>> isinstance(ctst_, CTST)
    True
    """

    actual: NDArray = field(repr=False)
    predicted: NDArray = field(repr=False)
    metric: Callable
    n_resamples: int = 9999
    rng: np.random.Generator = np.random.default_rng()
    n_jobs: int = 1
    batch: int | None = None
    alternative: Literal["less", "greater", "two_sided"] = "two_sided"

    def __post_init__(self):
        """Validate inputs."""
        self.actual = column_or_1d(self.actual)
        self.predicted = column_or_1d(self.predicted)
        check_consistent_length(self.actual, self.predicted)
        assert type_of_target(self.actual, "actual") == "binary"
        type_predicted = type_of_target(self.predicted, "predicted")
        assert type_predicted in (
            "binary",
            "continuous",
            "multiclass",
        ), f"Expected 'predicted' to be binary or continuous, got {type_predicted}."
        assert check_metric_function(self.metric), (
            f"'metric' expects a callable that conforms to scikit-learn metric. "
            f"{signature(self.metric)=} does not."
        )

    @cached_property
    def statistic(self) -> float:
        """
        Compute the observed test statistic.

        Returns
        -------
        float
            The test statistic.

        Notes
        -----
        The result is cached to avoid (expensive) recomputation.
        """
        return self.metric(self.actual, self.predicted)

    @cached_property
    def null(self) -> NDArray:
        """
        Compute the null distribution of the test statistic.

        Notes
        -----
        The result is cached to avoid (expensive) recomputation since the
        null distribution requires permutations.
        """
        return permutation_null(
            data=(self.actual, self.predicted),
            statistic=self.metric,
            n_resamples=self.n_resamples,
            rng=self.rng,
            n_jobs=self.n_jobs,
            batch=self.batch,
        ).astype(float)

    @cached_property
    def pvalue(self):
        """
        Compute the p-value using permutations.

        Notes
        -----
        The result is cached to avoid (expensive) recomputation.
        """
        test_ = EmpiricalPvalue(self.statistic, self.null)
        pvalue_ = methodcaller(self.alternative)(test_)
        return pvalue_

    @classmethod
    def from_samples(
        cls,
        first_sample: NDArray,
        second_sample: NDArray,
        metric: Callable,
        n_resamples: int = 9999,
        rng: np.random.Generator = np.random.default_rng(),
        n_jobs: int = 1,
        batch: int | None = None,
        alternative: Literal["less", "greater", "two_sided"] = "two_sided",
    ):
        """
        Create a CTST instance from two samples.

        Parameters
        ----------
        first_sample : NDArray
            First sample of scores. These can be binary or continuous.
        second_sample : NDArray
            Second sample of scores. These can be binary or continuous.

        Returns
        -------
        CTST
            An instance of the CTST class.
        """
        assert type_of_target(first_sample) == type_of_target(second_sample)
        samples = (first_sample, second_sample)
        actual = assign_labels(samples)
        predicted = concat_samples(samples)
        return cls(
            actual,
            predicted,
            metric,
            n_resamples,
            rng,
            n_jobs,
            batch,
            alternative,
        )


__all__ = [
    "CTST",
]



# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
from numpy.typing import NDArray


def assign_labels(samples: Sequence[NDArray]) -> NDArray:
    assert len(samples) > 1, f"{len(samples)=} must be greater than 1."
    labels = [np.repeat(i, np.array(s).shape[0]) for i, s in enumerate(samples)]
    return np.concatenate(labels, axis=None)


def group_by(data: NDArray, groups: NDArray):
    unique_groups, group_indices = np.unique(groups, return_inverse=True)
    grouped = [data[group_indices == i] for i in range(len(unique_groups))]
    return grouped


def concat_samples(samples: Sequence[NDArray]) -> NDArray:
    first_sample = next(iter(samples))
    if first_sample.ndim < 2:
        return np.concatenate(samples, axis=None)
    return np.concatenate(samples, axis=0)



# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
"""Helper functions to fit models in parallel."""

from __future__ import annotations

import os
from collections.abc import Iterable
from concurrent.futures import (
    FIRST_COMPLETED,
    Future,
    ThreadPoolExecutor,
    as_completed,
    wait,
)
from dataclasses import dataclass
from typing import Protocol, TypeVar, runtime_checkable

import numpy as np
from numpy.typing import NDArray

R = TypeVar("R", covariant=True)
C = TypeVar("C")


@dataclass
class Resampler:
    n_resamples: int = 9999
    rng: np.random.Generator = np.random.default_rng()
    batch: int | None = None
    n_jobs: int = 1

    def __post_init__(self):
        assert isinstance(self.rng, np.random.Generator)
        if self.n_jobs < 0:
            self.n_jobs = os.cpu_count() or 1
        assert self.n_resamples > 0 and isinstance(self.n_jobs, int)
        assert self.n_jobs > 0 and isinstance(self.n_jobs, int)
        if self.batch is None:
            self.batch = self.n_resamples
        assert self.n_resamples >= self.batch and isinstance(self.batch, int)

    def divide_work(self):
        batch_size, remainder = divmod(self.n_resamples, self.n_jobs)
        assert self.batch is not None, f"{self.batch=} must be an integer."
        if self.batch >= batch_size:
            yield from self._gen_work(self.n_jobs, batch_size, remainder)
        else:
            n_splits, remainder = divmod(self.n_resamples, self.batch)
            yield from self._gen_work(n_splits, self.batch, remainder)

    @classmethod
    def _gen_work(cls, n_iterations: int, batch_size: int, remainder: int):
        for _ in range(n_iterations):
            yield batch_size
        if remainder > 0:
            yield remainder


@runtime_checkable
class SupportsResampling(Protocol):
    def __call__(
        self,
        rng: np.random.Generator = np.random.default_rng(),
        n_resamples: int = 9999,
    ) -> NDArray[np.float64]: ...


def _simulate_in_parallel(
    simulate_fn: SupportsResampling,
    resampler: Resampler = Resampler(),
) -> NDArray[np.float64]:
    assert isinstance(simulate_fn, SupportsResampling)
    futures: set[Future] = set()
    results: Iterable[NDArray[np.float64]] = list()
    batch_sizes = list(resampler.divide_work())
    rngs = [rng for rng in resampler.rng.spawn(len(batch_sizes))]
    with ThreadPoolExecutor(max_workers=resampler.n_jobs) as executor:
        for rng, batch_size in zip(rngs, batch_sizes):
            if len(futures) >= resampler.n_jobs:
                completed, futures = wait(futures, return_when=FIRST_COMPLETED)
                for f in completed:
                    results.append(f.result())
            args_run = (rng, batch_size)
            futures.add(executor.submit(simulate_fn, *args_run))
        for f in as_completed(futures):
            results.append(f.result())
    return np.concat(results, axis=None)



# Copyright (c) 2025-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import Protocol, runtime_checkable

import numpy as np
from numpy.typing import NDArray

from samesame._parallel import Resampler, _simulate_in_parallel


@runtime_checkable
class SupportsDataPermutation(Protocol):
    def __call__(
        self,
        data: Sequence[NDArray],
        rng: np.random.Generator,
    ) -> Sequence[NDArray] | NDArray: ...


def _shuffle_first_column(
    data: Sequence[NDArray],
    rng: np.random.Generator,
):
    it = iter(data)
    first = next(it)
    rest = tuple(it)
    first_ = np.array(first, copy=True)
    first_ = rng.permutation(first_)
    return tuple((first_, *rest))


def _resample_null(
    data: Sequence[NDArray],
    statistic: Callable,
    permute: SupportsDataPermutation,
    n_resamples: int = 9999,
    rng: np.random.Generator = np.random.default_rng(),
) -> NDArray[np.float64]:
    results = np.empty(n_resamples, dtype=np.float64)
    for i in range(n_resamples):
        data_ = permute(data, rng)
        results[i] = statistic(*data_)
    return results


def permutation_null(
    data: Sequence[NDArray],
    statistic: Callable[[NDArray, NDArray], float],
    n_resamples: int = 9999,
    rng: np.random.Generator = np.random.default_rng(),
    n_jobs: int = 1,
    batch: int | None = None,
):
    resampler = Resampler(
        n_resamples=n_resamples,
        rng=rng,
        n_jobs=n_jobs,
        batch=batch,
    )
    def _resample_partial(rng=rng, n_resamples=n_resamples):
        return _resample_null(
            data=data,
            statistic=statistic,
            permute=_shuffle_first_column,
            n_resamples=n_resamples,
            rng=rng,
        )

    return _simulate_in_parallel(
        resampler=resampler,
        simulate_fn=_resample_partial,
    )
